""" removes initial indentation.

# Synopsys

  r= reindent(f'''
       def.....
          .....
       ''')

# Description

The indentation found in first line, is removed from all the other lines.

If newindent is provided, (' ' * newindent) is inserted in the begining of
    all the lines.

## Other functions ( printstr* = io.stringIO + reindent )

   1)   x = printstrinit("")    ## new strIO x
   2.1) printstr(x,"potatoes","and","other")  ## add to x
   2.3) printstrri(x, '''       
            def f ():
              ...
            ''')                ## add a reindented str to x
   3)   z = printstrval(x)      ## get the value and closeit
"""

import re
import io


def reindent(s, newindent=0):
    """ removes initial indentation.
    The indentation found in first line, is removed from all the other lines.

    If newindent is provided, (' ' * newindent) is inserted in the begining of
    all the lines.
"""

    s = re.sub(r"^(?=.)",r'\n',s)               # ensure initial \n
    indent = re.match(r'\n+([ \t]*)', s)[1]     # get indent of line1
    s = re.sub(rf"\n{indent}$",r'',s)           # rem final \nindent
    s = re.sub(rf"\n{indent}",rf'\n{" " * newindent}',s)  # reindent
    s = re.sub(rf"^\n",r'',s)                   # remove inicial empty line
    return s


def printstrinit(s="", newindent=0):
    output = io.StringIO()
    print(reindent(s,newindent), file=output)
    return output

def printstrri(output,s, newindent=0):
    print(reindent(s,newindent), file=output)

def printstr(output,*args, **kwargs):
    print(*args, file=output, **kwargs)

def printstrval(output):
    contents = output.getvalue()
    output.close()
    return contents



""" VERY EXPERIMENTAL :  module for opinied command line processing of OWL ontologies

# Synopsis

    jjtmd2ttl [option] file.t.md ...
      ## convert markdown with yaml metadate into turtle (seealso pandoc)

    jjtable2ttl [option] file.csv ...
      -F '#'            - Fiels separator (def: '::')
      -f ','            - sub field separatos (def: '[;,]')
      -h 'headerLine'   - set a "header line" (def: first line)
         Ex:  jjtable2ttl -F : -h 'Id::uid::gid::name::' /etc/passwd
      ## convert csv  metadate in turtle 

    jjyaml2ttl [option] file.yaml ...
       ... FIXME options descriptions missing
      ## converts multi-yaml dictionaries into turtle

    jjtax2ttl [option] file.tax ...
      -r 'isSonOf'      - subclass relation (def: 'rdfs:subClassOf')
      ## convert indented Class hierarchy in turtle

# Description
"""

from jjcli import *
#from unidecode import unidecode
import json
import yaml
from copy import deepcopy, copy
#import owlrl
#from   owlrl import convert_graph, RDFXML, TURTLE, JSON, AUTO, RDFA
#import rdflib
#from   rdflib.namespace import RDF, OWL, RDFS, SKOS, FOAF

##=======

## -----------------------------------------------------------------
## === utils importing "pseudo ontology elems"
##   parse tmd   ( meta(yaml) + body(markdown) )
##   tmd → turtle
##   t2iri(s)          """ term 2 IRIref """
##   t2iri_or_lit(s)
##   t2lit(s rel=p,)
##   docs2ttl(d, opt)       """ python structure to ttl

'''
FIXME:
 rdf:
    type          => s:iri o:iri.Class
    a             => s:iri o:iri.Class
    comment
 owl:
    inverseOf :            s:iri.rel → s:iri.rel
    Class
    Thing     :   iri.Class
    sameAs
    ObjectProperty :    if x 'a' ObjectProperty   /\ _x_ => s:iri o:iri
    DatatypeProperty  : if x 'a' DatatypeProperty /\ _x_ => s:iri o:literal
 rdfs:
    label :       s:iri       o:literal
    range :       s:iri.prop  o:iri.Class
    domain:       s:iri.prop  o:iri.Class
    subClassOf :  s:iri.Class o:iri.Class        
'''

###-----------------------------------------
### parse markdown with initial YAML-metadata  ( jjmd2ttl )
###-----------------------------------------

REL={'url': 'url'}  ## FIXME convensions

def md2py(c):
    """ files(markdown with initial yaml metadata) → docs = [ doc ] 
    inputfile=
       ---
       yaml-dict
       ---
       bodytext

    outputDoc=
       if bodytext not empty:
          doc["__body-format__"] = "markdown"
          doc[yaml[body] or "body"] = bodytext
       if -v:
          doc["__file__"] = inputfilename
"""
    docs =[]
    for txt in c.slurp():
        doc = {}
        ex = match(r'\s*---(.*?)---(.*)',txt,flags=re.S)
        if not ex:
            warn("### Erro: no formato de", c.filename())
            continue
        meta,text =(sub(r'\t','    ',ex[1]) , ex[2])
        try:
            doc = yaml.safe_load(meta)
        except Exception as e:
            warn("### Erro: nos metadados de", c.filename(),e)
            continue
        if not isinstance(doc,dict):
            warn("### Erro: nos metadados de", c.filename(),doc)
            continue
        if '-v' in c.opt:
            doc["__file__"] = c.filename()
        if search(r'\S',text):
            if "format" in doc:   
                bformat = doc.pop("format")
            else:
                bformat = "markdown"
            doc["__body-format__"] = bformat

            if "body" in doc:   
                bname = doc.pop("body")
            else:
                bname = "body"
            doc[f"{bname}^^{bformat}"] = text

        docs.append(doc)
    return docs

def main_tmd2ttl():            ## main -- tmd2ttl
    '''
NAME
    jjtmd2ttl - converts markdown with inicia YAML metadata into turtle
     -p properties-table (FIXME: ignored for the moment)

SYNOPSIS

Description
   inputfile=
    ---
    yaml-dict
    ---
    bodytext
'''
    c = clfilter(opt="do:",doc=main_tmd2ttl.__doc__)
    ds = md2py(c)
    print(docs2ttl(ds,c.opt))

#def quotes(s):
#    if not search(r'["\n]', str(s)):
#        return f'"{s.strip()}"
#    if search(r'[!?()"\'\n+,;/]',str(s)):
#        s1 = sub(r"\n",r" \n",s1)
#        return f'"""{s1} """'

###-----------------------------------------
### general IRI and lit
###-----------------------------------------

def t2lit(s,rel=None, ty=None):
    """ term 2 literal """
    if ty is not None: 
        typestr = f'^^:{ty}'
    elif rel in REL:
        typestr = f'^^:{REL[rel]}'
    else:
        typestr = ""

    if isinstance(s ,(dict,list,tuple)):
        return f'"""FIXME: {s}"""'
    if isinstance(s ,int):
        return str(s)
    if search(r'[!?()\'+,;/]',str(s)) and not search(r'["\r\n]', str(s)):
        return f'"{s.strip()}"{typestr}'
    if search(r'[!?()\'+,;/"\n\r]',str(s)):
        s1 = str(s).strip()
        if r := search(r'^!!(\w+)',s1):     ### !!id .... → """..."""^^:id
            s1 = sub(r'^!!\w+',r'',s1)
            return f'"""{s1} """^^:{str(r[1])}'
        else:
            s1 = sub(r"\n",r" \n",s1)
            return f'"""{s1} """{typestr}'

    return f'"{s.strip()}"{typestr}'

def t2iri(s, maybe=False):
    """ term 2 IRIref → IRI or "FIXME-..."None """
    if isinstance(s ,(dict,list,tuple)):
        warn("??? t2iri:",s)
        return f'"""FIXME: {s}"""'
    if maybe and match( r'\d', str(s)):
        return f'"""{str(s)}"""'
    if maybe and match( r'\d+$', str(s)):
        return f'"{str(s)}"'
    if s in {'type','a',} : return f'a'
    if s in {'inverseOf','Class','Thing','sameAs', 'ObjectProperty',
            'DatatypeProperty','Nothing'} :
        return f'owl:{s}'
    if s in {'range','domain','subClassOf'} :
        return f'rdfs:{s}'
    if search(r'^(owl|rdfs?|foaf|skos):\w+$',str(s)):
        return s

    if search(r'[!?()"\'\n+,;/ºª]',str(s)):
        aux = sub(r'[!?()"\'\n+,ºª;/ \r\t\n:_]+',r'_',s.strip()) 
        aux = aux.strip("_.,")
        if maybe: 
            return None
        else:
            return f":{aux}"

    iri = sub(r'[ \r\t\n:_]+',r'_',str(s).strip())
    iri = sub(r'\.$','',iri)
    return ":"+iri

def t2iri_or_lit(s,rel=None, ty=None):
    aux = t2iri(s,maybe=True)
    if aux is not None: 
        return aux
    else: 
        return t2lit(s,rel=rel,ty=ty)

###----------------------------------
### generate ttl
###----------------------------------

def getid(d, opt={}): ### { @id : vi, ...dict-tail } → {vi : ...dict-tail }
    if  not isinstance(d, dict):  return None
    subj = opt.get("-s")
    if   "id" in d:          return d.get("id")
    elif "ID" in d:          return d.get("ID")
    elif "@id" in d:         return d.get("@id")
    elif "name" in d:        return d.get("name")
    elif "@name" in d:       return d.get("@name")
    elif "title" in d:       return d.get("title")
    elif subj and subj in d: return d.get(subj)
    else:                    return None

def docs2ttl(d, opt={}):
    """ docs → turtle triples 
      (s,p,o)  or
      [s,p,o] → :s :p :o.      
           or more exactly dict2ttl({ s : {p : o}})

      [s, dict({pi:oi})] → :s :p1 :o1 ... :s :pi :oi .
           or more exactly dict2ttl({ s : dict(...)})

      list → join( "\n", [docst2ttl(v) v ∈ list] )

      dict → dict2ttl(dict)
    """

    if isinstance(d,list):
        if len(d)==3 and not isinstance(d[0],(list,dict,tuple)) :
            return dict2ttl({d[0]: {d[1]: d[2]}}, opt)
        elif ( len(d)==2 and
                not isinstance(d[0],(list,dict,tuple)) and
                isinstance(d[1],dict) ) :
            return dict2ttl({d[0]: d[1]}, opt)
        else:
            return str.join("\n",[docs2ttl(x,opt) for x in d])
    if isinstance(d,dict):
        return dict2ttl(d, opt)
    if isinstance(d,tuple):
        if len(d)==3:
            return dict2ttl({d[0]: {d[1]: d[2]}}, opt)
        else:
            warn("????",d)

def dict2ttl(d:dict,opt={}):
    """ dict doc → turtle triples 
    
  explicit subject:(@id | @name | id | ID | name| title) in dict
     { @id : vi, ...dict-tail } →⇒   dict2ttl {vi : ...dict-tail }
     Ex: { "@id" : "João", 
                 "parent" : [ "D", "M"], 
                 "a" : "Person" }
       → (:João :parent :D)
         (:João :parent :M)
         (:João a :Person)

     { Manel : { pred : obj }}    
       → (:Manel :pred :ob)
     { Manel : { pred : setObs }} 
       → {(:Manel :pred :ob) | ob ∈ setObs}

     Ex: { "Manel": { "parent" : [ "D", "M"], "son": ["A","L"] }, 
                "Joana": { "a": "Person"} }
       → (:Manel :parent :D) 
         (:Manel :parent :M) 
         (:Manel :son :A) 
         (:Manel :son :L) 
         (:Joana a :Person)

     { "ont": val }  ⇒  doc2ttl(val)
     Ex:  { "@id": "Rui", "a": "Person", "ont": [ (a2,a3,a4),(a3,a4,{a4,a5,a6})]  },
       → (:rui a :Person)
         (:a2 :a3 :a4)
         (:a3 :a4 :a4)
         (:a3 :a4 :a5)
         (:a3 :a4 :a6)
"""
    r = ""
    rd = deepcopy(d)
    subj = opt.get("-s")
    ### { @id : vi, ...dict-tail } → {vi : ...dict-tail }
    if   "id" in d:          s = rd.pop("id");    rd = {s: rd}
    elif "ID" in d:          s = rd.pop("ID");    rd = {s: rd}
    elif "@id" in d:         s = rd.pop("@id");   rd = {s: rd}
    elif "name" in d:        s = rd.pop("name");  rd = {s: rd}
    elif "@name" in d:       s = rd.pop("@name"); rd = {s: rd}
    elif subj and subj in d: s = rd.pop(subj); rd = {s: rd}
#    else:                    s = f"FIXME2-{d.keys()}"; rd = {s: rd}

    for s,dd in rd.items():
        if match(r'@?ont(ology|ologia)?$',s,flags=re.I):
            r += docs2ttl(dd, opt)
            continue

        if not isinstance(dd,dict):
            warn("Error: expecting a dictionary, got a ",rd)
            dd = {"__DEBUG__": dd}

        for p,o in dd.items():
            if match(r'@?ont(ology|ologia)?$',p,flags=re.I):
                r += docs2ttl(o, opt)
            elif isinstance(o,(list,set)):
                for oo in o:
                    if isinstance(oo,dict):
                        r += dict2ttl({s: {p: oo}}, opt)
                    else:
                        r += tripttl(s,p,oo)
            elif isinstance(o,dict):
                for k in o:
                    r += tripttl(s,p,k)
                r += dict2ttl(o, opt)
            else:
                r += tripttl(s,p,o)
    return r

def tripttl(s, p, o, force=None, oty=None ):
    s1 = t2iri(s) 
    if pp := search(r'(.*)\^\^:?(.+)', p):
        oty = pp[2]
        p1 = t2iri(pp[1]) 
        force="lit"
    else:
        p1 = t2iri(p) 
       
    if   force is  None: o1 = t2iri_or_lit(o, rel=p, ty=oty)
    elif force == "lit": o1 = t2lit(o, rel=p, ty=oty)
    elif force == "iri": o1 = t2iri(o)
    return f'{s1} {p1} {o1} .\n'

###-----------------------------------------
### parse table (jjtable2ttl)
###-----------------------------------------


def parse_head(h: str, opt ):
    fs  = opt.get('-F','::')    ## fiel separator (def: "::")
    fs2 = ':'
    idclass,*ftits = split(fr'\s*{fs}\s*',h.strip())
    for i,f in enumerate(ftits):
        a = match(fr'(.+){fs2}(.+)',f)
        if a :
            ftits[i] = {"rel": a[1], "class": a[2]}
        else:
            ftits[i] = {"rel":f}
    return (idclass,ftits)

def parse_tup(t:str,idclass,ftits,opt ):
    fs  = opt.get('-F','::')
    fs2 = opt.get('-f','[;,|]')
    ont =[]
    id,*fields = split(fr'\s*{fs}\s*',t.strip())
    doc = {}
    for i,f in enumerate(fields):
        if not search(r'\S',f): 
            continue
        if i >= len(ftits):
            warn('To many fields ',id,fields)
            continue
        else:
            finf = ftits[i]
        if not search(r'\S',finf["rel"]): 
            continue
        a = [ x for x in split(fr'\s*{fs2}\s*',f) if search(r'\S',x) ]
        doc[finf["rel"]] = a
        if "class" in finf :
            for campo in a :
                ont.append( (campo, "type", finf["class"]))
    if ont:
        doc["ont"] = ont
    doc["@id"]= id
    doc["type"]= doc["type"]+[idclass] if "type" in doc else idclass
    return doc

def table2py(c):
    """ files(headerline + CSV(fs="::",fs2="[;,|]") → docs = [ doc ] """
    headopt = c.opt.get("-h")  ## ex: table2tt -h 'Person :: pai:Person :: f'
    docs =[]
    for txt in c.text():
        tabs = split(r'\s*@TAB\s*', txt)
        for tab in tabs:
            if not search(r'\w',tab):
                continue

            if headopt:
                head = headopt
                tups = [x for x in split(r'\s*\n\s*',tab) if match(r'\w',x)]
            else:
                head,*tups = [x for x in split(r'\s*\n\s*',tab) if match(r'\w',x)]

            idclass,ftits= parse_head(head,c.opt)
            for line in tups:
                doc= parse_tup(line,idclass,ftits,c.opt)
                if '-v' in c.opt:
                    doc["__file__"] = c.filename()
                docs.append(doc)

    return docs

def main_table2ttl():
    """
NAME
   jjtable2ttl - converts a "::" field separated table in turtle

SYNOPIS
   jjtable2tt [option] files > out.ttl
     Options:
       -F "="    field setarator "=" (def: "::")
       -f ":"    subfield separator ":" (def: [,;|])
       -h "Person :: son:Person :: year" define an external header

Example
   FIXME: incomplete
"""
    c = clfilter(opt="h:do:F:f:v", doc=main_table2ttl.__doc__) 
    ds = table2py(c)
    print(docs2ttl(ds,c.opt))

###-----------------------------------------
### parse a multi-yaml file: (jjyaml2ttl)
###-----------------------------------------

def yamls2py(c,files=None):
    """ files( ("---" yaml)+ ) → docs = [ doc ] """
    docs =[]
    
    for txt in c.slurp(files):
        isa_serie = "-c" in c.opt
        txt = sub(r'\t','    ',txt)  ## repl tabs (they are invalid in yaml)
        txt = sub(r'(\n\w+:)(\w)',r'\1 \2',txt)  ## (abc:def → abc: def)
        count = 0
        the_1 = None

        for txtd in re.split(r'\n---\s',txt):
            try:
                doc = yaml.safe_load(txtd)
            except Exception as e:
                warn("### Erro yaml de", c.filename(),e)
                continue
            count += 1  
            if not doc:
                continue
            if not isinstance(doc,dict):
                warn("### Erro invalid format:", c.filename(),doc, txtd )
                continue
            if the_1 is None :
                if isa_serie or doc.get("type") in {"serie", "list"}: ##.FIXME
                    isa_serie = True
                    the_1 = doc
                    parent = getid(the_1, c.opt)
                the_1 = doc
            else:
                if isa_serie:
                    if parent:
                        doc["pof"] = parent
                    doc["__index__"] = count

            if "-v" in c.opt:
                doc["__file__"] = f"{c.filename()}--{count}"
            docs.append(doc)
    return docs
    '''
        try:
            elems = yaml.safe_load_all(txt)
        except Exception as e:
            warn("### Erro yaml de", c.filename(),e)
            continue
        if not elems:
            continue
        count = 0
        for doc in elems: 
            count += 1  
            if not doc:
                continue
            if "-v" in c.opt:
                doc["__file__"] = f"{c.filename()}--{count}"
            if not isinstance(doc,dict):
                warn("### Erro invalid format:", c.filename(),doc)
                continue
            docs.append(doc)
    return docs
    '''

def main_yaml2ttl():
    """
NAME
   jjyaml2ttl - prints a turtle ontology from a (multi-doc)YAML

SYNOPSIS
   jjyaml2ttl [options] files
     -v       verbose

     -s nnn   nnn is the subject field id. (def: @id | @name | id | name)
     -c       multi-doc (first doc is the container; others are the parts)
             for i, x in enumerate(docs[1:])
                 subj(x) :pof subj(first doc)
                 subj(x) :_index_  i

     -p properties-table (FIXME: ignored for the moment)


Description
   yamldocs → turtle triples
   1) if YAMLdoc is tuple or list with the following shape:

      (s,p,o)  or [s,p,o] → :s :p :o.
           or more exactly⇒ dict2ttl({ s : {p : o}})

      [s, dict({pi:oi})] → :s :p1 :o1 ... :s :pi :oi .
           or more exactly⇒ dict2ttl({ s : dict(...)})

      list ⇒  join( "\n", [doc2ttl(v) v ∈ list] )

      dict ⇒  dict2ttl(dict)

   2) if YAMLdoc is a dict: dict → turtle triples 
    
      if doc has a explicit subject name: 
         ( @id | @name | id | ID | name) in dict
      
         { @id : vi, dict-tail } ⇒ dict2ttl {vi : dict-tail }
      Ex1:  { "@id" : "João", 
              "parent" : [ "D", "M"], 
              "a" : "Person" }
       → (:João :parent :D)
         (:João :parent :M)
         (:João a :Person)

      Ex2: {Manel : { pred : obj }}    
       → (:Manel :pred :ob)

      Ex3: {Manel : { pred : setObs }} 
      → { (:Manel :pred :ob) | ob ∈ setObs}
 
      Ex4: {Manel: {parent :[D, M], son: [A,L] }, 
            Joana: {a: Person} }
      → (:Manel :parent :D) 
        (:Manel :parent :M) 
        (:Manel :son :A) 
        (:Manel :son :L) 
        (:Joana a :Person)

      If doc has a "ont"|"ontology" key
      • { "ont": val }  ⇒  doc2ttl(val)
      Ex5: { "@id": "Rui", "a": "Person", "ont": [ (a2,a3,a4),(a3,a4,{a4,a5,a6})]  },
       → (:rui a :Person)
         (:a2 :a3 :a4)
         (:a3 :a4 :a4)
         (:a3 :a4 :a5)
         (:a3 :a4 :a6)
"""
    c = clfilter(opt ="do:F:f:vp:cs:",doc=main_yaml2ttl.__doc__)     
    ds = yamls2py(c)
    print(docs2ttl(ds, c.opt))

###-----------------------------------------
### Taxonomy
###-----------------------------------------

def FIXME_taxstr2py(c, rel="rdfs:subClassOf"):
    """ str ( ????  ) → docs = [ doc ] """
    top_name = "owl:Thing"
    inda = 0
    tabind = [-1]
    ant = [top_name]    

    for line in c.splitlines():
        if not search(r'\w',line): continue 

        if m := match(r'\s*#', line):
            print(line)
            continue
        if m := match(r'([ .-]*)(\S.*)', line):
            term = m[2] 
            ind = len(m[1])
        if ind <= inda :
            while len(tabind) and ind <= tabind[-1]:
                tabind.pop()
                ant.pop()
        tabind.append(ind)
        print( tripttl(term, rel, ant[-1], "iri" ),end="")
        ant.append(term) 
        inda = ind

def tax2py(c, rel="rdfs:subClassOf",files=None):
    """ files( ????  ) → docs = [ doc ] """
    if "-r" in c.opt:
        rel = c.opt["-r"]

    top_name = c.opt.get("-t", ":Thing")
    inda = 0
    tabind = [-1]
    ant = [top_name]    

    for line in c.input(files):
        if not search(r'\w',line): continue 

        if m := match(r'\s*#', line):
            print(line)
            continue
        if m := match(r'([ .-]*)(\S.*)', line):
            term = m[2] 
            ind = len(m[1])
        if ind <= inda :
            while len(tabind) and ind <= tabind[-1]:
                tabind.pop()
                ant.pop()
        tabind.append(ind)
        print( tripttl(term, rel, ant[-1], "iri" ),end="")
        ant.append(term) 
        inda = ind

def main_tax2ttl():
    """
NAME
    jjtax2ttl - converter indented taxonomy to turtle classes

SYNOPSIS
    jjtax2ttl [option] file.tax ...
      -r 'isSonOf'   - name of the subclass relation (def: 'rdfs:subClassOf')

Description
    Converts indented Class hierarchy in turtle. 
    Example 

     science
       physics
          dynamics
    #     statics       ...comment: removed for the moment 
       math
          algebra
          trignometry
     doc on literature

"""

    c = clfilter(opt ="do:r:t:v",doc=main_tax2ttl.__doc__)    
    tax2py(c)
#    ds = tax2py(c)
#    print(docs2ttl(ds, c.opt))



#!/usr/bin/python3
""" VERY EXPERIMENTAL :  module for opinied command line processing of OWL ontologies

# Synopsis

    owlcat  a.ttl b.owl c.ttl > all.ttl
      -t                - input in Turtle (def: based on extension)
      -f xml            - define output Format
      -n                - skip inference of new triples (def: infer)
      -o file
      ## concatenate → infer new triples → removeprefix

    owlclass a.ttl      
      -t                - input in Turtle (def: based on extension)
      -r                - keep transitive Redundances
      ## show class taxonomy

    owlgrep pattern a.ttl
      -t                - input in turtle (def: based on extension)
      -k                - just the IRI (default: term and adjacents)
       pattern :
         intanceRE
         classRE::instanceRE
      ## grep indivituals, class, or class:indi; ouput their triples

    owllabel2term a.ttl > new.ttl   
      -s nameSpace                  - default #
      ## rename IRIRef from rfds:label

    owlexpandpp a.ttl   - expand and pretty print
      -t                - input in turtle (def: based on extension)
      ##

    owlxdxf  a.ttl      
      -n dict_name      - title of the generated dictionary
      -t                - input in turtle (def: based on extension)
      -o out.xdxf       - redirects output (default stdout)
      ## Build a XDXF linked dictionary (seealso goldendict)

    jjtmd2ttl [option] file.t.md ...
      ## convert markdown with yaml metadate in turtle (seealso pandoc)

    jjyaml2ttl [option] file.t.yaml ...
      ## convert multi-yaml in turtle 

    jjtable2ttl [option] file.csv ...
      -F '#'            - Fiels separator (def: '::')
      -f ','            - sub field separatos (def: '[;,]')
      -h 'headerLine'   - set a "header line" (def: first line)
         Ex:   table2ttl -F : -h 'Id::uid::gid::name::' /etc/passwd
      ## convert csv  metadate in turtle 

    jjtax2ttl [option] file.tax ...
      -r 'isSonOf'      - subclass relation (def: 'rdfs:subClassOf')
      ## convert indented Class hierarky in turtle 

As a module:

    import jjowl
    ... FIXME

# Description
"""

__version__ = "0.2.3"

from jjcli import *
from jjowl.reindent import *
import os
from unidecode import unidecode
import json
import yaml
import owlrl
from   owlrl import convert_graph   ## , RDFXML, TURTLE, JSON, AUTO, RDFA
import rdflib
from   rdflib.namespace import RDF, OWL, RDFS, SKOS, FOAF, Namespace


def OFF_best_name_d(g : rdflib.Graph) -> dict :
   """ FIXME: IRIRef -> str based on RDFS.label, id, ??? """
   bestname={}
   for n in g.all_nodes():
       if islit(n): continue
       if name := g.preferredLabel(n) :
           bestname[s]=name[0].strip()
       else:
           txt = term.n3(g.namespace_manager)
           txt = sub(r'^:', '', txt)
           txt = sub(r'_', ' ', txt)
           bestname[s]=txt.strip()
   return bestname
#   for s,p,o in g.triples((None,RDFS.label,None)):
#   return


def get_invs(g: rdflib.Graph) -> dict :
   """Dictionary of inverse relations (based on inverseOf and SymetricProperty"""
   inv = {OWL.sameAs: OWL.sameAs}
   for s,p,o in g.triples((None,OWL.inverseOf,None)):
       inv[s]=o
       inv[o]=s
   for s,p,o in g.triples((None,RDF.type, OWL.SymmetricProperty)):
       inv[s]=s
   return inv

def reduce_graph(g,fixuri=None,fixlit=None)-> rdflib.Graph:
   def fix(item):
        if islit(item) and fixlit:
            return rdflib.term.Literal(fixlit(str(item)))
        if isinstance(item, rdflib.term.URIRef) and fixuri:
            return rdflib.term.URIRef(fixuri(str(item)))
        return item

   fixed = rdflib.Graph()
   fixed.bind('owl', OWL)
   fixed.bind('rdf', RDF)
   fixed.bind('rdfs', RDFS)
   fixed.bind('skos', SKOS)
   fixed.bind('foaf', FOAF)

   fixed += [ (fix(s), fix(p), fix(o)) for s,p,o in g]
   return fixed

def concatload(files:list, opt: dict) -> rdflib.Graph :
   ns = opt.get("-s",'#')
   g = rdflib.Graph(base=ns)
   g.bind("",rdflib.Namespace(ns))
   g.namespace_manager.bind('owl', OWL)
   g.bind('rdf', RDF)
   g.bind('rdfs', RDFS)
   g.bind('skos', SKOS)
   g.bind('foaf', FOAF)
   for f in files:
      if ".n3" in f or ".ttl" in f or "-t" in opt:
         try:
             g.parse(f,format='n3')    ## more flexible than turtle
         except Exception as e:
             warn("#### Error in ", f,e)
      else:
         try:
             g.parse(f)                ## format='xml'
         except Exception as e:
             warn("#### Error in ", f,e)
   return g

def concat(files:list, opt: dict) -> rdflib.Graph :
   ns=opt.get("-s",'#')
   g=concatload(files, opt)
   def fixu(t):
      if str(RDF) in t or str(OWL) in t or str(RDFS) in t :
          return t
      else:
          return  sub(r'(http|file).*[#/]',ns,t)

   g2=reduce_graph(g, fixuri=fixu)
   g2.bind("",ns)
   return g2

def termcmp(t):
   return  unidecode(sub(r'.*[#/]','',t).lower())

def islit(t):
   return isinstance(t,rdflib.term.Literal)

###----------------------
### main entry points
###----------------------

def mcat():           # main for owlcat
   '''
    # Synopsis

    owlcat  a.ttl b.owl c.ttl > all.ttl
      -t          - input in Turtle (def: based on extension)
      -f xml      - define output Format (def: turtle)
      -n          - skip inference of new triples (def: infer)
      -o file
      -s namepace    - namespace for ":"    (def: #)

    # description

      ## concatenate → infer new triples → removeprefix
   '''

   c=clfilter(opt="i:f:no:s:t" , doc=mcat.__doc__)
   g=owlproc(c.args,c.opt)
   g.serial()

def mlabel2term():    # main for owllabel
   '''
   '''
   c=clfilter(opt="f:kpcno:s:t", doc=mlabel2term.__doc__)
   g=owlproc(c.args,c.opt)
   g.rename_label2term()
   g.serial()

def mgrep():          # main for owlgrep
   '''
   # Name: owlgrep

    owlgrep pattern a.ttl
      -t                - input in turtle (def: based on extension)
      -k                - just the IRI (default: term and adjacents)
       pattern :
         intanceRE
         classRE::instanceRE

    # Description

     grep inputpattern: individuals, class, or class:indi; 
          ouput: their triples
   '''
   c=clfilter(opt="f:kpco:s:t", doc=mgrep.__doc__)
   pat=c.args.pop(0)
   g=owlproc(c.args,c.opt)
   if v:= match(r'(.+?)::(.+)',pat):
       g.grep2(v[1],v[2])
   else:
       g.grep(pat)

def mexpandpp():       # main for owlexpand
   c=clfilter(opt="kpco:s:t")
   g=owlproc(c.args,c.opt)
   g.pprint()

def mxdxf():           # main for owlxdxf
   '''
   '''
   c=clfilter(opt="b:no:s:t", doc=mxdxf.__doc__)
   if "-b" not in c.opt:
       basedir = os.path.dirname(os.path.realpath(c.args[0]))
       c.opt["-b"]=basedir
   g=owlproc(c.args,c.opt)
   g.xdxf()

def mhtml():           # main for owlhtml
   '''
   '''
   c=clfilter(opt="b:no:s:t", doc=mhtml.__doc__)
   if "-b" not in c.opt:
       basedir = os.path.dirname(os.path.realpath(c.args[0]))
       c.opt["-b"]=basedir
   g=owlproc(c.args,c.opt)
   g.html()

def mclass():          # main for owlclass
   '''
    owlclass a.ttl      
      -t                - input in Turtle (def: based on extension)
      -r                - keep transitive Redundances
      ## show class taxonomy
   '''
   c=clfilter(opt="no:rs:t", doc=mclass.__doc__)
   g=owlproc(c.args,c.opt)
   g.pptopdown(OWL.Thing)

def mprops():          # main for owlprop
   '''
   '''
   c=clfilter(opt="no:rs:t", doc=mprops.__doc__)
   g=owlproc(c.args,c.opt)
   g.ppprops()

###-----------------
### owlproc
###-----------------

class owlproc:
   """ Class for process owl ontologies
      .opt
      .g
      .inv
      .instances :  tipo -> [indiv]
      .subcl   : class -> {class}
      .supcl   : class -> {class}
      .supcltc : class -> {class}

   """
   def __init__(self,ontos,
                     opt={},
                     ):
       self.opt=opt
       if "-s" not in self.opt:    ## default name space
           self.ns='#'
       else:
           self.ns=self.opt["-s"]

       self.g=concat(ontos,opt)
       if "-n" not in self.opt:
           self._infer()
       self.inv=get_invs(self.g)
       self._get_subclass()
       self._instances()
       self.IT = Namespace(self.ns)

   def serial(self,fmt="turtle"): ## serializer to fmt (def:turtle)
       if "-f" in self.opt :
           fmt = self.opt["-f"]
       print(self.g.serialize(format=fmt))

   def _pp(self,term) -> str:
       """ returns a Prety printed a URIRef or Literal """
       return term.n3(self.g.namespace_manager)

   def _instances(self):
       self.instances={}
       for s,p,o in self.g.triples((None,RDF.type, None)):
           self.instances[o]=self.instances.get(o,[]) + [s]

   def _infer(self):
       owlrl.DeductiveClosure(owlrl.OWLRL_Extension_Trimming ).expand(self.g)

###------------------
### owl-props
###------------------

   def ppprops(self):
      predbag={}
      predtyp={}
      for s in self.g.subjects(predicate=RDF.type,object=OWL.ObjectProperty): 
          predbag[s]=0
          predtyp[s]="ObjectProperty"
      for s in self.g.subjects(predicate=RDF.type,object=OWL.DatatypeProperty): 
          predbag[s]=0
          predtyp[s]="DatatypeProperty"
      for p in self.g.predicates():
          predbag[p]=predbag.get(p,0)+1 
      for p in predbag.keys():
          for s,o in self.g.subject_objects(p):   ## ↓ p o
              if islit(o):
                 predtyp[p]="DatatypeProperty"
              else:
                 predtyp[p]="ObjectProperty"
              break
      for p,n in sorted(predbag.items()):
          print(f'{self._pp(p)} a {predtyp[p]} . # {n} ')


###---------------------
### XDXF
###---------------------
#  p.xdxf()               // ONT → xdxf-file
#  p._xpp(term) → str     // a xdxf Pretyprinted URIRef
#  p._term_inf_xdxf(n,f): // ONT , 

   def xdxf(self):
      if "-o" in self.opt:
          f = open(self.opt["-o"],"w",encoding="utf-8")
      else:
          f = sys.stdout
      ignore_pred={ RDF.type, RDFS.subPropertyOf , OWL.topObjectProperty,
         OWL.equivalentProperty }

      if "-n" in self.opt:
          title=self.opt["-n"]
      else:
          title="Dicionário"

      print( reindent( f"""
          <?xml version="1.0" encoding="UTF-8" ?>
          <xdxf lang_from="POR" lang_to="DE" format="logical" revision="033">
              <meta_info>
                  <title>{title}</title>
                  <full_title>{title}</full_title>
                  <file_ver>001</file_ver>
                  <creation_date></creation_date>
              </meta_info>
          <lexicon>
         """), file=f)

      for n in sorted(self.g.all_nodes(), key=termcmp):
          if islit(n): continue
          if n in [OWL.Thing, OWL.Nothing, RDFS.Class] : continue
          self._term_inf_xdxf(n,f)  ## FIXME
      print("</lexicon>\n</xdxf>",file=f)
      if "-o" in self.opt:
          f.close()

   def _term_inf_xdxf(self,n,f):
      ignore_pred={ RDF.type, RDFS.subPropertyOf , OWL.topObjectProperty,
         OWL.equivalentProperty }

      print("",file=f)
      print(f'<ar><k>{self._xpp(n)}</k><def>',file=f) ####  AR
      cls = self._simplify_classes(self.g.objects(subject=n,predicate=RDF.type),
                                 strat="bu")
      for c in cls:                                     ## class
          print( f"<kref>{self._xpp(c)}</kref>",
                 file=f)
      for p,o in sorted(self.g.predicate_objects(n)):   ## ↓ p o
          if p in ignore_pred: continue
          ppstr=self._pp(p)
          opstr=self._pp(o)
          pstr=self._xpp(p)
          ostr=self._xpp(o)
          if ppstr in {':img','IMG',':jpg',':png'}:
              print(f"   <def><img width='500px' src={opstr}/></def>",
                    file=f)
          elif ppstr in {':pdf',':url',':midi'}:
              if "-b" in self.opt:
                  aux = opstr.strip('"')
                  basedir = self.opt["-b"]
                  url = f'"file:///{basedir}/{aux}"'
              print(f"   <def><iref href={url}>{pstr} ↗</iref></def>",
                    file=f)
          elif islit(o):
              print(f"   <def>{pstr}: {ostr}</def>",
                    file=f)
          else:
              print(f"   <def>{pstr}: <kref>{ostr}</kref></def>",
                    file=f)

      for s,p in sorted(self.g.subject_predicates(n)):   ## s p ↓
          if p in ignore_pred  or p in self.inv: continue
          print(f"   <def><kref>{self._xpp(s)}</kref>  {self._xpp(p)} *</def>",
                file=f)
      if n in self.instances:
          for i in sorted(self.instances[n],key=termcmp):
              print(f" <def><kref>{self._xpp(i)}</kref></def>",
                    file=f)
      print(f'</def></ar>', file=f)                    ####  /AR


   def _xpp(self,term) -> str:
       """ returns a xdxf Prety printed URIRef """
       txt = self._pp(term)
       txt = sub(r'\\"','"',txt)
       if islit(term):
           txt = sub(r'[<>&]','§',txt)
           txt = sub(r'§(/?)(sup|def|br|kref|b|i|iref|img)\b(.*?)§',r'<\1\2\3>',txt)
           txt = txt.strip("""'" \t\n""")
           if '"""' in txt  or "'''" in txt:
               return "<deftext>"+ txt + "</deftext>"
           elif match(r'(.+)\.(jpe?g|png|pdf|svg)$', txt):
               return txt
           else:
               return txt
       else:
           txt = txt.strip("""'" \t\n""")
           txt = sub(r'^:', '', txt)
           txt = sub(r'_', ' ', txt)
           txt = sub(r'[<>&]','§',txt)
           txt = sub(r'§(/?)(sup|def|br|kref|b|i|iref|img)\b(.*?)§',r'<\1\2\3>',txt)
           return txt

###---------------------
### grep
###---------------------

   def grep(self,pat):
       for n in sorted(self.g.all_nodes(), key=termcmp):
           if islit(n): continue
           if n in [OWL.Thing, OWL.Nothing, RDFS.Class] : continue
           npp = self._pp(n)
           if search( pat, npp, flags=re.I):
               if "-k" in self.opt:
                   print(npp)
               else:
                   self._pterm_inf(n)

   def grep2(self,patc,pati):
       for s,o in self.g.subject_objects(RDF.type):
           opp = self._pp(o)
           if search( patc, opp, flags=re.I):
               spp = self._pp(s)
               if search( pati, spp, flags=re.I):
                   if "-k" in self.opt:
                       print(f'{opp}::{spp}')
                   else:
                       self._pterm_inf(s)

###---------------------
### classes of the ontology
###---------------------

   def _get_subclass(self):
      self.subcl   = {}
      self.supcl= {OWL.Thing: set()}
      self.supcltc = {}
      for s,p,o in self.g.triples((None,RDF.type,None)):
          self.subcl.setdefault(o,set())
          self.supcl.setdefault(o,set())
      for s,p,o in self.g.triples((None,RDFS.subClassOf,None)): # s < o
          self.subcl.setdefault(s,set())
          self.subcl.setdefault(o,set())
          self.supcl.setdefault(s,set())
          self.supcl.setdefault(o,set())
          self.supcl[s].add(o)
          self.subcl[o].add(s) 

      roots= set()
      for x,s in self.supcl.items():
          if x == OWL.Thing: continue
          if s == set():
              roots.add(x)
              self.supcl[x].add(OWL.Thing)
          if s == { OWL.Thing }:
              roots.add(x)
      self.subcl[OWL.Thing] = roots

      self.mksupcltc(OWL.Thing)
      self._simplify_subcl()
      self._simplify_supcl()

   def mksupcltc(self, top, vis={}, acu=set() ):
       if top in vis: return
       vis[top] = 1
       self.supcltc[top] = acu | self.supcl[top]
       for child in self.subcl.get(top,[]):
           self.mksupcltc( child, vis, acu=self.supcltc[top] )

   def pptopdown(self,top,vis={},indent=0,max=1000):
       if max <= 0  : return
       if top in vis: return
       vis[top]=1
       print( f'{"  " * indent}{self._pp(top)}' )
       scls=self.subcl.get(top,[])
#       if "-r" not in self.opt:
#           scls=self._simplify_classes(self.subcl.get(top,[]))
       for a in sorted(scls,key=termcmp):
           self.pptopdown(a,vis,indent+2,max-1)

   def _topdown(self,top,vis={},max=1000):
       if max == 1  : return self._pp(top)
       if max <= 0  : return None
       if top in vis: return None
       vis[top]=1
       scls=self.subcl.get(top, set())
#       scls=self._simplify_classes(scls)
       childs=[]
       if childs is []: 
           return (self._pp(top), )
       else:
           for a in sorted(scls,key=termcmp):
               childs.append(self._topdown(a,vis,max-1))
           return (self._pp(top), childs)

   def _simplify_subcl(self):
       for c1, c2s in self.subcl.items():
           aux = c2s - { OWL.NamedIndividual, OWL.Nothing, RDFS.Class}
           for x in aux.copy():
               ys = self.supcltc.get(x,set())
               if len(ys & aux) > 0 :
                   aux.remove(x)
           self.subcl[c1]=aux

   def _simplify_supcl(self):
       for c1, c2s in self.supcl.items():
           aux = c2s - { OWL.NamedIndividual, OWL.Nothing, RDFS.Class}
           for x in aux.copy():
               ys = self.supcltc.get(x,set())
               aux -= ys
           self.supcl[c1]=aux

   def _simplify_classes(self, cls:list, strat="td") -> list:
       """ FIXME: remove redundant class from a class list"""
       aux = set(cls) - {OWL.Thing, OWL.NamedIndividual, OWL.Nothing, RDFS.Class}
       for x in aux.copy():
           ys = self.supcltc.get(x,set())
           if strat == "td":
               if len(ys & aux) > 0 :
                   aux.remove(x)
           else:    ## "bu"
               aux -= ys
       return aux

###----------------
### label to 
###----------------

   def rename_label2term(self) -> rdflib.Graph :
       newname = {}

       for s,o in self.g.subject_objects(RDFS.label):
           base = sub(r'(.*[#/]).*',r'\1',str(s))
           newid = sub(r' ','_', str(o).strip())
           newname[str(s)] = base + newid

       for s,o in self.g.subject_objects(SKOS.prefLabel):
           base = sub(r'(.*[#/]).*',r'\1',str(s))
           newid = sub(r' ','_', str(o).strip())
           newname[str(s)] = base + newid

       def rename(t):
          if str(RDF) in t or str(OWL) in t or str(RDFS) in t or str(SKOS) in t:
              return t
          else:
              taux = newname.get(t,t)
              return sub(r'(http|file).*[#/]',self.ns,taux)
       g2=reduce_graph(self.g, fixuri=rename)
       g2.bind("",self.ns)
       self.g = g2

   def _pterm_inf(self,n,opt={}):
       ignore_pred={ RDF.type, RDFS.subPropertyOf , OWL.topObjectProperty,
          OWL.equivalentProperty }

       print("====")
       print(self._pp(n))
       cls = self._simplify_classes(self.g.objects(subject=n,predicate=RDF.type),
                                  strat="bu")
       for c in cls:
           print("   ", self._pp(c))

       prevpred=""
       for p,o in sorted(self.g.predicate_objects(n)):   ## ↓ p o
           p_pp = self._pp(p)
           if p_pp != prevpred: 
               print( "       ", p_pp)
               prevpred = p_pp
           print( "           ", self._pp(o))

#       for p,o in sorted(self.g.predicate_objects(n)):   ## ↓ p o
#           if p in ignore_pred: continue
#           print( "       ", self._pp(p), self._pp(o))

       for s,p in sorted(self.g.subject_predicates(n)):   ## s p ↓
           if p in ignore_pred: continue
           if p in self.inv: continue
           print( "       ", self._pp(s), self._pp(p), "*")

   def pprint(self):
       for n in sorted(self.g.all_nodes(), key=termcmp):
           if islit(n): continue
           if n in [OWL.Thing, OWL.Nothing, RDFS.Class] : continue
           self._pterm_inf(n)

###---------------------
### HTML
###---------------------

   def html(self):
      if "-o" in self.opt:
          f = open(self.opt["-o"],"w",encoding="utf-8")
      else:
          f = sys.stdout
      ignore_pred={ RDF.type, RDFS.subPropertyOf , OWL.topObjectProperty,
         OWL.equivalentProperty }

      if "-n" in self.opt:
          title=self.opt["-n"]
      else:
          title="Dicionário"

      print(f"""<html><head> <title>{title}</title>""",
          reindent("""
              <meta charset="UTF-8"/>
              <style>
                 section {
                   background-color: #ffffff;
                   margin: 10px;
                   border: solid 1px #999999;
                   padding: 5px;
                   clear:both;
                 }

                 b[lang] { color: red; }

                 img { object-fit: contain; }

                 /* img.r { float: right; }
                 img.r::after { content: ""; clear: both; display: table;
                 } */

                 a { color: #000099; text-decoration: none; }
                 a:hover { color: #0000ff; text-decoration: underline;  }
                 
                 ul { list-style-type: none; }
                 li.lingua strong { color: #009900;}
                 li.lingua a { font-style: italic; /* text-transform: lowercase; */ }
                 li.lingua { margin-bottom: 2px; } 

                 li.rel strong  { color: #999900; }
                 li.text strong { color: #009999; }
                 li.text        { padding-top: 10px; padding-bottom: 10px; }
              </style>
              <script src="https://cdn.jsdelivr.net/npm/marked/marked.min.js"></script>
              </head>
              <body>

              <input type="text" id="myInput" onkeyup="myFunction()" 
                     placeholder="Search for names.." title="Type in a name"/>

              """), file=f)

      for n in sorted(self.g.all_nodes(), key=termcmp):
          if islit(n): continue
          if n in [OWL.Thing, OWL.Nothing, RDFS.Class] : continue
          self._term_inf_html(n,f)  ## FIXME

      print(reindent("""
          <script>
             function myFunction() {
                var input, filter, sections, term, i, txtValue;
                input = document.getElementById("myInput");
                filter = input.value.toUpperCase();
                sections = document.getElementsByTagName("section");
                for (i = 0; i < sections.length; i++) {
                   term = sections[i].getElementsByTagName("strong")[0];
                   if (term) {
                      txtValue = term.textContent || term.innerText;
                      if (txtValue.toUpperCase().indexOf(filter) > -1) {
                         sections[i].style.display = "";
                      } else {
                         sections[i].style.display = "none";
                      }
                   }
                }
             }
             
             function ensurevisid(a) {
                sec = document.getElementById(a);
                if(sec){ sec.style.display=""; }
             }
             
             marks = document.getElementsByClassName("markdown");
             for (i = 0; i < marks.length; i++) {
                marks[i].innerHTML = marked.parse( marks[i].innerHTML );
             }
 
          </script>
          </body>\n</html>
          """), file=f)
      if "-o" in self.opt:
          f.close()


   def _term_inf_html(self,n,f):
      ignore_pred={ RDF.type, RDFS.subPropertyOf , OWL.topObjectProperty,
         OWL.equivalentProperty }

      print(reindent(f"""
          <section id='{self._hpp(n)}'>
          <strong>{self._hpp(n)}</strong>
          <ul>

          """), file=f)
      cls = self._simplify_classes(self.g.objects(subject=n,predicate=RDF.type),
                                 strat="bu")
#      cls = self.supcl.get(n,set())
      print("<li>",file=f)
      for c in cls:                                     ## class
          print( f"<a href='#{self._hpp(c)}'>{self._hpp(c)}</a>", file=f)
      print("</li>",file=f)

      print("<ol>",file=f)
      for p,o in sorted(self.g.predicate_objects(n)):   ## ↓ p o
          if p in ignore_pred: continue
          ppstr = self._pp(p)
          opstr = self._pp(o)
          pstr = self._hpp(p)
          ostr = self._hpp(o)
          if ppstr in {':img','IMG',':jpg',':png'}:
              print(f"   <li><img width='500px' src={opstr}/></li>",
                    file=f)
          elif ppstr in {':pdf',':url',':midi'}:
              if "-b" in self.opt:
                  aux = opstr.strip('"')
              print(f"   <li><a href='{aux}'>{pstr} ↗</a></li>",
                    file=f)
          elif islit(o):
              print(f"   <li>{pstr}: {ostr}</li>",
                    file=f)
          else:
              print(f"   <li>{pstr}: <a href='#{ostr}'>{ostr}</a></li>",
                    file=f)

      print("</ol>",file=f)
      for s,p in sorted(self.g.subject_predicates(n)):   ## s p ↓
          if p in ignore_pred  or p in self.inv: continue
          print(f"   <li><a href='#{self._hpp(s)}'>{self._hpp(s)}</a>  {self._hpp(p)} *</li>",
                file=f)
      if n in self.instances:
          print("<li>Instances <ol>",file=f)
          for i in sorted(self.instances[n],key=termcmp):
              print(f" <li><a href='#{self._hpp(i)}'>{self._hpp(i)}</a></li>",
                    file=f)
          print("</ol></li>",file=f)
      print(f'</ul></section>', file=f)                    ####  /AR


   def _hpp(self,term) -> str:
       """ returns a html-Pretty-printed URIRef """
       txt = self._pp(term)
       txt = sub(r'\\"','"',txt)
       if islit(term):
           txt = sub(r'[<>&]','§',txt)
           txt = sub(r'§(/?)(sup|def|br|a|b|i|p|iref|img)\b(.*?)§',r'<\1\2\3>',txt,flags=re.I)
           txt = txt.strip("""'" \t\n""")
           if '"""' in txt  or "'''" in txt:
               return "<div>"+ txt + "</div>"
           elif match(r'(.+)\.(jpe?g|png|pdf|svg)$', txt):
               return txt
           else:
               return txt
       else:
           txt = txt.strip("""'" \t\n""")
           txt = sub(r'^:', '', txt)
           txt = sub(r'_', ' ', txt)
           txt = sub(r'[<>&]','§',txt)
           txt = sub(r'§(/?)(sup|def|br|a|b|i|p|iref|img)\b(.*?)§',r'<\1\2\3>',txt)
           return txt

